/**
 * ADT skup realizovan otvorenih hashovanjem
 * 
 * Modifikovana verzija za ilustracije i test program
 */
public class OHashSet<T> implements Set<T> {

	private static final int DEFAULT_TABLE_SIZE = 101;

	private static class Node {
		Object value;
		Node next;

		public Node(Object value) {
			this.value = value;
		}
	}

	private Node[] table;

	private int count;
	private int[] tableCount;

	public OHashSet() {
		this(DEFAULT_TABLE_SIZE);
	}

	public OHashSet(int size) {
		if (size <= 0)
			throw new IllegalArgumentException(
					"Velicina hash tabele ne moze biti negativna");

		table = new Node[size];
		tableCount = new int[size];
		count = 0;
	}

	private int hash(T o) {
		if (o == null)
			throw new IllegalArgumentException(
					"Hash funkcija se ne moze racunati za null objekat");

		return Math.abs(o.hashCode() % table.length);
	}

	/**
	 * Pretrazuje listu kolizija za datu vrednost hash funkcije (drugi argument
	 * metoda) trazeci element (prvi argument metoda).
	 * 
	 * Metod vraca null ukoliko element ne postoji u tabeli, ili niz od dve
	 * reference: 1. prva referenca je referenca na trazeni element 2. druga
	 * referenca je referenca na element ispred trazenog elementa
	 */
	private Node[] searchColissionChain(T element, int hashValue) {
		Node current = table[hashValue];
		Node prev = null;

		while (current != null) {
			if (current.value.equals(element)) {
				Node[] ret = new Node[2];
				ret[0] = current;
				ret[1] = prev;
				return ret;
			}

			prev = current;
			current = current.next;
		}

		return null;
	}

	public boolean add(T element) {
		int hashValue = hash(element);
		Node[] n = searchColissionChain(element, hashValue);
		if (n != null) {
			return false;
		}

		Node newElement = new Node(element);
		newElement.next = table[hashValue];
		table[hashValue] = newElement;

		count++;
		tableCount[hashValue]++;
		return true;
	}

	public boolean remove(T element) {
		int hashValue = hash(element);
		Node[] n = searchColissionChain(element, hashValue);
		if (n == null)
			return false;

		if (n[0] == table[hashValue]) {
			table[hashValue] = table[hashValue].next;
		} else {
			n[1].next = n[0].next;
		}

		count--;
		tableCount[hashValue]--;
		return true;
	}

	public boolean contains(T element) {
		return searchColissionChain(element, hash(element)) != null;
	}

	public void print() {
		for (int i = 0; i < table.length; i++) {
			System.out.print("HashCode = " + i + ": ");

			Node head = table[i];
			if (head == null) {
				System.out.println(" empty");
			} else {
				while (head != null) {
					System.out.print("[" + head.value + "]");
					head = head.next;
				}
				System.out.println();
			}
		}
	}

	public String toString() {
		return "OHashSet; br elemenata " + getCount();
	}

	public int getCount() {
		return count;
	}

	public int size() {
		return getCount();
	}

	public double getChiSquare() {
		double chi = 0.0;

		if (count > 0 && table.length > 0) {
			double avg = (double) count / table.length;
			for (int i = 0; i < tableCount.length; i++) {
				chi += (tableCount[i] - avg) * (tableCount[i] - avg) / avg;
			}
		}
		return chi;
	}

	public double getPercentFull() {
		if (tableCount.length == 0) {
			return 0;
		}

		int filled = 0;

		for (int i = 0; i < tableCount.length; i++) {
			if (tableCount[i] > 0)
				filled++;
		}

		return filled * 100.0 / tableCount.length;
	}

	public void printStats() {
		System.out.printf("Elements    : %3d%n", getCount());
		System.out.printf("Percent full: %6.2f %n", getPercentFull());
		System.out.printf("Chi square  : %8.4f %n", getChiSquare());
		System.out.printf("'Columns'   : %3d%n", table.length);
	}

	@SuppressWarnings("unchecked")
	public T someElement() {
		if (count > 0) {
			for (int i = 0; i < table.length; i++) {
				Node head = table[i];
				if (head != null) {
					return (T) head.value;
				}
			}
		}
		return null;
	}
}
